import platform
import re
from munch import Munch
from . import SolarisMountEntry
from ...base.mount import MountRepositoryMixin

from logging import getLogger
log = getLogger()

class SolarisMountRepositoryMixin(MountRepositoryMixin):
    def _get_mount_entry_class(self):
        return SolarisMountEntry

    def _read_fstab(self):
        return self._read_file("/etc/vfstab")

    def _read_mtab(self):
        return self._read_file("/etc/mnttab")

    def _translate_empty_values(self, entries):
        keys_to_none_string = ["dirname"]
        for entry in entries:
            for key in keys_to_none_string:
                if entry.has_key(key) and entry[key] == '-':
                    entry[key] = "none"

    def _parse_options_for_entry(self, entry):
        string = entry["opts"]
        results = {}
        pattern = re.compile(OPTION_PATTERN)
        for match in pattern.finditer(string):
            key = match.groupdict().get("key")
            value = match.groupdict().get("value")
            if not key == '-':
                results[key] = self._translate_value(value)
        entry["opts"] = results

    def _get_list_of_groupdicts_from_mtab(self):
        pattern = re.compile(MOUNT_ENTRY_PATTERN_MTAB_SOLARIS, re.MULTILINE)
        string = self._read_mtab()
        log.debug("mtab content = \n{}".format(string))
        results = [match.groupdict() for match in pattern.finditer(string)]
        self._translate_empty_values(results)
        return self._parse_options_in_entries(results)

    def _get_list_of_groupdicts_from_fstab(self):
        pattern = re.compile(MOUNT_ENTRY_PATTERN_FSTAB_SOLARIS, re.MULTILINE)
        string = self._read_fstab()
        log.debug("fstab content = \n{}".format(string))
        results = [match.groupdict() for match in pattern.finditer(string)]
        self._translate_empty_values(results)
        return self._parse_options_in_entries(results)

WORD_PATTERN = r"[^# \t\n\r\f\v]+"
FSNAME_PATTERN = r"(?P<fsname>{})".format(WORD_PATTERN)
DIRNAME_PATTERN = r"(?P<dirname>{})".format(WORD_PATTERN)
TYPNAME_PATTERN = r"(?P<typename>{})".format(WORD_PATTERN)
STRING_PATTERN = r"[^,=# \t\n\r\f\v]+"
OPTION_PATTERN = r"(?P<key>{})(?:=(?P<value>{}))?".format(STRING_PATTERN, STRING_PATTERN)
OPTS_PATTERN = r"(?P<opts>{})".format(WORD_PATTERN)
FREQ_PATTERN = r"(?P<freq>\d*)"
PASSNO_PATTERN = r"(?P<passno>[\-\d]*)"
CREATIONTIME_PATTERN = r"(?P<creationtime>\d*)"
SEP = r"[ \t]+"

MOUNT_ENTRY_PATTERN_MTAB_SOLARIS = r"^{fsname}{sep}{dirname}{sep}{typename}{sep}{opts}{sep}{creationtime}$".format(
                                           sep=SEP,
                                           fsname=FSNAME_PATTERN,
                                           dirname=DIRNAME_PATTERN,
                                           typename=TYPNAME_PATTERN,
                                           opts=OPTS_PATTERN,
                                           creationtime=CREATIONTIME_PATTERN)
MOUNT_ENTRY_PATTERN_FSTAB_SOLARIS = r"^{fsname}{sep}{unparsed}{sep}{dirname}{sep}{typename}{sep}{passno}{sep}{mountonboot}{sep}{opts}$".format(
                                           sep=SEP,
                                           unparsed=WORD_PATTERN,
                                           fsname=FSNAME_PATTERN,
                                           dirname=DIRNAME_PATTERN,
                                           typename=TYPNAME_PATTERN,
                                           passno=PASSNO_PATTERN,
                                           mountonboot=WORD_PATTERN,
                                           opts=OPTS_PATTERN)



__import__("pkg_resources").declare_namespace(__name__)

from ..base import MountManagerBaseClass

from .mount import SolarisMountRepositoryMixin
from .supported_fs import SolarisSupportedFileSystemsMixin
from .mounter import SolarisMounterMixin

class SolarisMountManager(MountManagerBaseClass, SolarisMountRepositoryMixin, SolarisSupportedFileSystemsMixin, SolarisMounterMixin):
    pass



import platform
import re
from munch import Munch
from ...base.mount import MountRepositoryMixin

from logging import getLogger
log = getLogger()

class LinuxMountRepositoryMixin(MountRepositoryMixin):
    def _read_fstab(self):
        return self._read_file("/etc/fstab")

    def _read_mtab(self):
        return self._read_file("/etc/mtab")

    def _read_utab(self):
        return self._read_file("/run/mount/utab")

    def _parse_options_for_entry(self, entry):
        string = entry["opts"]
        results = {}
        pattern = re.compile(OPTION_PATTERN)
        for match in pattern.finditer(string):
            key = match.groupdict().get("key")
            value = match.groupdict().get("value")
            results[key] = self._translate_value(value)
        entry["opts"] = results
        return results

    def _get_entries_dict_from_utab(self):
        pattern = re.compile(UTAB_ENTRY_PATTERN_LINUX, re.MULTILINE)
        string = self._read_utab()
        log.debug(u"utab content = \n{}".format(string))
        results = dict([(match.groups()[0], self._parse_options_for_entry(match.groupdict())) for match in
                       pattern.finditer(string)])
        return results

    def _get_list_of_groupdicts_from_mtab(self):
        pattern = re.compile(MOUNT_ENTRY_PATTERN_LINUX, re.MULTILINE)
        string = self._read_mtab()
        log.debug(u"mtab content = \n{}".format(string))
        utab_results = self._get_entries_dict_from_utab()
        mtab_results = [match.groupdict() for match in pattern.finditer(string)]
        mtab_results = self._parse_options_in_entries(mtab_results)
        for mtab_result in mtab_results:
            fsname = mtab_result['fsname']
            if fsname in utab_results:
                mtab_result['opts'].update(utab_results[fsname])

        return mtab_results

    def _get_list_of_groupdicts_from_fstab(self):
        pattern = re.compile(MOUNT_ENTRY_PATTERN_LINUX, re.MULTILINE)
        string = self._read_fstab()
        log.debug(u"fstab content = \n{}".format(string))
        results = [match.groupdict() for match in pattern.finditer(string)]
        return self._parse_options_in_entries(results)


WORD_PATTERN = r"[^# \t\n\r\f\v]+"
FSNAME_PATTERN = r"(?P<fsname>{})".format(WORD_PATTERN)
DIRNAME_PATTERN = r"(?P<dirname>{})".format(WORD_PATTERN)
TYPNAME_PATTERN = r"(?P<typename>{})".format(WORD_PATTERN)
STRING_PATTERN = r"[^,=# \t\n\r\f\v]+"
OPTION_PATTERN = r"(?P<key>{})(?:=(?P<value>{}))?".format(STRING_PATTERN, STRING_PATTERN)
OPTS_PATTERN = r"(?P<opts>{})".format(WORD_PATTERN)
FREQ_PATTERN = r"(?P<freq>\d*)"
PASSNO_PATTERN = r"(?P<passno>[\-\d]*)"
SEP = r"[ \t]+"

MOUNT_ENTRY_PATTERN_LINUX = r"^{fsname}{sep}{dirname}{sep}{typename}{sep}{opts}{sep}{freq}{sep}{passno}$".format(
                                           sep=SEP,
                                           fsname=FSNAME_PATTERN,
                                           dirname=DIRNAME_PATTERN,
                                           typename=TYPNAME_PATTERN,
                                           opts=OPTS_PATTERN,
                                           freq=FREQ_PATTERN,
                                           passno=PASSNO_PATTERN)

UTAB_ENTRY_PATTERN_LINUX = r"^SRC={fsname}{sep}TARGET={dirname}{sep}ROOT=/{sep}OPTS={opts}$".format(
                                           sep=SEP,
                                           fsname=FSNAME_PATTERN,
                                           dirname=DIRNAME_PATTERN,
                                           opts=OPTS_PATTERN)



from ..base.mounter import MounterMixin, execute_mount

class LinuxMounterMixin(MounterMixin):
    def _get_entry_format(self, entry):
        return str(entry)

    def mount_entry(self, entry):
        args = ["-t", entry.get_typename(), entry.get_fsname(), entry.get_dirname()]
        args.extend(self._format_options(entry))
        execute_mount(args)



from munch import Munch


class MountEntry(object):

    DIRNAME_REPLACE_COUPLES = (
       (r"\040", " "),     # HIP-688 spaces appear different in mtab
       (r"\134", "\\"),    # HPT-1905
    )

    @classmethod
    def from_groupdict(cls, groupdict):
        return cls(**groupdict)

    def __init__(self, fsname, dirname, typename, opts=dict(), freq=0, passno=0):
        self._bunch = Munch(fsname=fsname, dirname=dirname,
                            typename=typename, opts=opts,
                            freq=freq, passno=passno)

    def get_fsname(self):
        """:returns: name of mounted file system"""
        return self._bunch.fsname

    def get_dirname(self):
        """:returns: file system path prefix"""
        replaced_dirname = self._bunch.dirname
        for original_value, replaced_value in self.DIRNAME_REPLACE_COUPLES:
            replaced_dirname = replaced_dirname.replace(original_value, replaced_value)
        return replaced_dirname

    def get_typename(self):
        """:returns: mount type"""
        return self._bunch.typename

    def get_opts(self):
        """:returns: mount options"""
        return self._bunch.opts

    def get_freq(self):
        """:returns: dump frequency in days"""
        return self._bunch.freq

    def get_passno(self):
        """:returns: pass number on parallel fsck"""
        return self._bunch.passno

    def _str_options(self):
        if self.get_opts().keys() == []:
            return ''
        options = ''
        for key, value in self.get_opts().items():
            if value is True:
                options += "{},".format(key)
            else:
                options += "{}={},".format(key, value)
        return options.strip(',')

    def __str__(self):
        return "\t".join(item.encode("utf-8") for item in
                         [self.get_fsname(), self.get_dirname(), self.get_typename(),
                         self._str_options(), self.get_freq(), self.get_passno()])



from ..base.supported_fs import SupportedFileSystemsMixin

class LinuxSupportedFileSystemsMixin(SupportedFileSystemsMixin):
    def get_supported_file_systems(self):
        """:returns: a list of mountable filesystems on this host
        :rtype: a list of strings"""
        result = set()
        for name in self._get_list_of_internally_supported_file_systems() + \
                    self._get_list_of_supported_file_systems_by_helpers():
            result.add(name)
        return [name for name in result]

    def get_creatable_file_systems(self):
        from os import listdir
        mountable = self.get_supported_file_systems()
        creatable = [fs[5:] for fs in listdir("/sbin") if fs.startswith("mkfs.")]
        creatable_and_mountable = [fs for fs in mountable if fs in creatable]
        return creatable_and_mountable

    def _is_ext4_supported(self):
        # HIP-1035
        # some RedHat-based distributions are 'techonology preview', and don't
        # have exte4 listed in /etc/filesystems although it exists inside
        # some other versions of RedHat have a bug that ext4 is not listed
        # in /etc/filesystems
        # so this is a workground for these old versions of RedHat and variants
        from os import path
        return path.exists(path.join(path.sep, '/sbin/mkfs.ext4'))

    def _get_list_of_internally_supported_file_systems(self):
        # mount will try to read the file /etc/filesystems,  or,
        # if  that  does  not  exist, /proc/filesystems.  All of the filesystem types listed there will be
        # tried, except for those that are labeled "nodev" (e.g., devpts, proc and nfs).  If /etc/filesys-
        # tems ends in a line with a single * only, mount will read /proc/filesystems afterwards.
        result = set()
        for name in self._get_filesystem_names_from_etc():
            result.add(name)
        if len(result) == 0 or (self._get_etc_filesystems().splitlines() and
                                self._get_etc_filesystems().splitlines()[-1] == '*'):
            for name in self._get_filesystem_names_from_proc():
                result.add(name)
        if self._is_ext4_supported():
            result.add('ext4')
        return [name for name in result]

    def _get_filesystem_names_from_etc(self):
        result = []
        for line in self._get_etc_filesystems().splitlines():
            if line == '' or line.startswith("nodev") or '*' in line:
                continue
            result.append(line)
        return result

    def _get_filesystem_names_from_proc(self):
        result = []
        for line in self._get_proc_filesystems().splitlines():
            if line == '' or line.startswith("nodev"):
                continue
            result.append(line.strip())
        return result

    def _get_proc_filesystems(self):
        from os.path import exists
        path = "/proc/filesystems"
        if not exists(path):
            return ''
        with open(path) as fd:
            return fd.read()

    def _get_etc_filesystems(self):
        from os.path import exists
        path = "/etc/filesystems"
        if not exists(path):
            return ''
        with open(path) as fd:
            return fd.read()

    def _get_list_of_supported_file_systems_by_helpers(self):
        from glob import glob
        basedir = "/sbin"
        replace_string = "{}/mount.".format(basedir)
        return map(lambda path: path.replace(replace_string, ""),
                   glob("{}*".format(replace_string)))




from munch import Munch
from ...base.mount import MountEntry

class SolarisMountEntry(MountEntry):
    def __init__(self, fsname, dirname, typename, opts=dict(), freq=0, passno='-', mountonboot="yes", creationtime=0):
        super(SolarisMountEntry, self).__init__(fsname, dirname, typename, opts, freq, passno)
        self._bunch.update(Munch(mountonboot=mountonboot, creationtime=creationtime))

    def _str_options(self):
        if self.get_opts().keys() == []:
            return ''
        options = ''
        for key, value in self.get_opts().items():
            if value is True:
                options += "{},".format(key)
            else:
                options += "{}={},".format(key, value)
        return options.strip(',')

    def __str__(self):
        string_double_sep = "\t\t".join(str(item) for item in \
                        [self.get_fsname(), '-', self.get_dirname().replace("none", '-'), self.get_typename()])
        string_one_sep = "\t".join(str(item) for item in \
                        [self.get_passno(), self.get_mountonboot(), self._str_options().replace('', '-')])
        return "\t".join([string_double_sep, string_one_sep])



__import__("pkg_resources").declare_namespace(__name__)

from .mount import MountEntry, MountRepositoryMixin
from .supported_fs import SupportedFileSystemsMixin
from .mounter import MounterMixin

class MountManagerBaseClass(object):
    pass

class MountManager(MountManagerBaseClass, MountRepositoryMixin, SupportedFileSystemsMixin, MounterMixin):
    pass



from ..base.mounter import MounterMixin, execute_mount

class SolarisMounterMixin(MounterMixin):
    def _get_fstab_path(self):
        return "/etc/fstab"

    def _get_entry_format(self, entry):
        return entry.get_format_solaris()

    def mount_entry(self, entry):
        args = ["-F", entry.get_typename(), entry.get_fsname(), entry.get_dirname()]
        args.extend(self._format_options(entry))
        execute_mount(args)



from ctypes import CDLL, create_string_buffer
from ..base.supported_fs import SupportedFileSystemsMixin

LIBC_SO_NAME = "libc.so"
SYSFC_GET_NAME_OPCODE = 2
SYSFC_COUNT_OPCODE = 3
SYSFC_NAME_SIZE = 20

CREATABLE_FS = ["ufs", "zfs", "fat", "fat32", "vfat"]

class SolarisSupportedFileSystemsMixin(SupportedFileSystemsMixin):
    def get_supported_file_systems(self):
        """:returns: a list of mountable filesystems on this host
        :rtype: a list of strings"""
        result = set()
        for name in self._get_filesystem_names_from_libc() + \
                    self._get_list_of_supported_file_systems_by_helpers():
            result.add(name)
        return [name for name in result]

    def get_creatable_file_systems(self):
        supported_fs = self.get_supported_file_systems()
        result = []
        for fs in CREATABLE_FS:
            if fs in supported_fs:
                result.append(fs)
        return result

    def _get_filesystem_names_from_libc(self):
        libc = CDLL(LIBC_SO_NAME)
        fscount = libc.sysfs(SYSFC_COUNT_OPCODE)
        fsname = create_string_buffer(SYSFC_NAME_SIZE)
        result = []
        for i in xrange(1, fscount + 1):
            if not (libc.sysfs(SYSFC_GET_NAME_OPCODE, i, fsname) == -1 or fsname.value == ''):
                result.append(fsname.value)
        return result

    def _get_list_of_supported_file_systems_by_helpers(self):
        from glob import glob
        basedir = "/usr/lib/fs"
        replace_string = "{}/".format(basedir)
        return map(lambda path: path.replace(replace_string, ""),
                   glob("{}*".format(replace_string)))



__import__("pkg_resources").declare_namespace(__name__)

__all__ = ['get_mount_manager']

from logging import getLogger
from infi.exceptools import chain
from infi.os_info import get_platform_string
from infi.pyutils.lazy import cached_function

logger = getLogger(__name__)

def _get_platform_specific_mountmanager_class():
    from .base import MountManager as PlatformMountManager
    from brownie.importing import import_string
    plat = get_platform_string().split('-')[0]
    platform_module_string = "{}.{}".format(__name__, plat)
    platform_module = import_string(platform_module_string)
    try:
        PlatformMountManager = getattr(platform_module, "{}MountManager".format(plat.capitalize()))
    except AttributeError:
        msg = "Failed to import platform-specific mount manager"
        logger.exception(msg)
        raise chain(ImportError(msg))
    return PlatformMountManager

@cached_function
def get_mount_manager():
    """returns a global instance of a `infi.mount_utils.base.MountManager`. """
    return _get_platform_specific_mountmanager_class()()



__import__("pkg_resources").declare_namespace(__name__)

from ..base import MountManagerBaseClass

from .mount import LinuxMountRepositoryMixin
from .supported_fs import LinuxSupportedFileSystemsMixin
from .mounter import LinuxMounterMixin

class LinuxMountManager(MountManagerBaseClass, LinuxMountRepositoryMixin, LinuxSupportedFileSystemsMixin, LinuxMounterMixin):
    pass



from infi.exceptools import InfiException

class MountException(InfiException):
    pass

class IncorrectInvocationOrPermissions(MountException):
    pass

class SystemErrorException(MountException):
    pass

class MountInternalBugException(MountException):
    pass

class UserInterruptException(MountException):
    pass

class ProblemWithWritingOrLockingException(MountException):
    pass

class MountFailureException(MountException):
    pass

class SomeMountSucceededException(MountException):
    pass


ERRORCODES_DICT = {1:IncorrectInvocationOrPermissions,
                   2:SystemErrorException,
                   4:MountInternalBugException,
                   8:UserInterruptException,
                   16:ProblemWithWritingOrLockingException,
                   32:MountFailureException,
                   64:SomeMountSucceededException}

def translate_mount_error(errorno):
    if ERRORCODES_DICT.has_key(errorno):
        return ERRORCODES_DICT.get(errorno)
    else:
        return MountException(errorno)



class SupportedFileSystemsMixin(object):
    def get_supported_file_systems(self):
        """:returns: a list of mountable filesystems on this host
        :rtype: a list of strings"""
        raise NotImplemented()

    def get_creatable_file_systems(self):
        raise NotImplemented()



from ..exceptions import translate_mount_error
from infi.exceptools import chain
from infi.pyutils.contexts import contextmanager

def execute(commandname, args):
    from logging import getLogger
    log = getLogger()
    debug = log.debug

    from infi.execute import execute as _execute
    debug(u'executing {} {}'.format(commandname, ' '.join(args)))
    subprocess = _execute([commandname] + args)
    debug('waiting for it')
    subprocess.wait()
    debug('returncode = %s', subprocess.get_returncode())
    debug('stdout = %s', subprocess.get_stdout())
    debug('stderr = %s', subprocess.get_stderr())
    if subprocess.get_returncode() != 0:
        raise chain(translate_mount_error(subprocess.get_returncode())())

def execute_mount(args):
    execute("mount", args)

def execute_umount(args):
    execute("umount", args)

class MounterMixin(object):
    def _format_options(self, entry):
        if entry.get_opts().keys() == []:
            return []
        options = ''
        for key, value in entry.get_opts().items():
            if value is True:
                options += "{},".format(key)
            else:
                options += "{}={},".format(key, value)
        return ["-o", options.strip(',')]

    def _get_fstab_path(self):
        return "/etc/fstab"

    @contextmanager
    def _get_fstab_context(self, mode='a'):
        from os.path import exists
        if not exists(self._get_fstab_path()):
            raise IOError()
        with open(self._get_fstab_path(), mode) as fd:
            yield fd

    def _read_fstab(self):
        with open(self._get_fstab_path(), 'r') as fd:
            return fd.read()

    def _get_entry_format(self, entry):
        raise NotImplementedError()

    def mount_entry(self, entry):
        raise NotImplementedError()

    def umount_entry(self, entry):
        args = [entry.get_dirname(), ]
        execute_umount(args)

    def add_entry_to_fstab(self, entry):
        with self._get_fstab_context() as fd:
            fd.write("{}\n".format(self._get_entry_format(entry)))

    def remove_entry_from_fstab(self, entry):
        content = self._read_fstab()
        line_to_remove = self._get_entry_format(entry).strip()
        with self._get_fstab_context('w') as fd:
            for line in [line.strip() for line in content.splitlines()]:
                if line_to_remove == line:
                    continue
                fd.write("{}\n".format(line))



import platform
import re
from munch import Munch
from . import MountEntry

from logging import getLogger
log = getLogger()

class MountRepositoryMixin(object):
    def _get_mount_entry_class(self):
        return MountEntry

    def _read_file(self, path):
        from os.path import exists
        if not exists(path):
            return ''
        with open(path) as fd:
            return fd.read().decode("utf-8")

    def _translate_value(self, value):
        if value is None:
            return True
        if value.isdigit():
            return int(value)
        return value

    def _parse_options_for_entry(self, entry):
        raise NotImplementedError()

    def _parse_options_in_entries(self, entries):
        for entry in entries:
            self._parse_options_for_entry(entry)
        return entries

    def _get_entries_dict_from_utab(self):
        raise NotImplementedError()

    def _get_list_of_groupdicts_from_mtab(self):
        raise NotImplementedError()

    def _get_list_of_groupdicts_from_fstab(self):
        raise NotImplementedError()

    def get_mounts_from_mtab(self):
        return [self._get_mount_entry_class().from_groupdict(groupdict) for groupdict in self._get_list_of_groupdicts_from_mtab()]

    def get_mounts_from_fstab(self):
        return [self._get_mount_entry_class().from_groupdict(groupdict) for groupdict in self._get_list_of_groupdicts_from_fstab()]

    def is_path_mounted(self, dirname):
        return any(item.get_dirname() == dirname for item in self.get_mounts_from_mtab())

    def is_fs_mounted(self, fsname):
        return any(item.get_fsname() == fsname for item in self.get_mounts_from_mtab())

    def is_entry_in_fstab(self, entry):
        for item in self.get_mounts_from_fstab():
            if item.get_fsname() == entry.get_fsname() and item.get_dirname() == entry.get_dirname():
                return True
        return False
